#ifndef __INTIMAGE_H__
#define __INTIMAGE_H__

#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <math.h>

#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

#include "mdarray.h"

const REAL PI = (REAL)3.14159265358979323846;
const REAL PI4 = (REAL)(PI/4);

template<class T>
class IntImage:public Array2dC<T>
{
private:
    IntImage(const IntImage<T> &source) { } // prohibit copy constructor

public:
    IntImage():variance(0.0),label(-1) { }
    ~IntImage() { Clear(); }

    inline void Clear(void);
    inline void SetSize(const int h, const int w);
    void Print(std::ostream& of = std::cout) const;
    bool Load(const std::string& filename);
    void Save(const std::string& filename) const;
    void Swap(IntImage<T>& image2);

    void Resize(IntImage<T> &result,const REAL ratio) const;
    void Resize(IntImage<T>& result,const int height,const int width) const;
	IplImage* convertImageRGBtoHSV(const IplImage *imageRGB);

	IntImage<T>& operator=(const IntImage<T>& source);
    void operator-=(const IntImage<T>& img2);
    void AbsoluteValue(void);
    void Thresh(const REAL thresh);

public:
    using Array2dC<T>::nrow;
    using Array2dC<T>::ncol;
    using Array2dC<T>::buf;
    using Array2dC<T>::p;
    REAL variance;
    int label;
};

template<class T>
void IntImage<T>::Clear(void)
{
    Array2dC<T>::Clear();
    variance = 0.0;
    label = -1;
}

template<class T>
bool IntImage<T>::Load(const std::string& filename)
{
    IplImage* img;
    IplImage* img2;
	
    img = cvLoadImage(filename.c_str());
	img2=convertImageRGBtoHSV(img);
    if(img==NULL) return false;

	CvScalar imageOriPix;							
	int **Hue=new int* [img->height];
	int **Saturation=new int* [img->height];
	int **Value=new int* [img->height];	
	int **Color=new int* [img->height];

	for ( int i = 0 ; i < img->height ; i++ ){		
		Hue[i]=new int[img->width];	
		Saturation[i]=new int[img->width];
		Value[i]=new int[img->width];
		Color[i]=new int[img->width];
		for ( int j=0 ; j<img->width ; j++ ){		
			Hue[i][j]=0;
			Saturation[i][j]=0;	
			Value[i][j]=0;
			Color[i][j]=0;
		}											
	}															
	for(int i=0;i<img->height;i++){
		for(int j=0;j<img->width;j++){
			imageOriPix = cvGet2D(img2,i,j);
			Hue[i][j] = (int)imageOriPix.val[0];
			Saturation[i][j] = (int)imageOriPix.val[1];
			Value[i][j] = (int)imageOriPix.val[2];
			Hue[i][j]=(Hue[i][j]/128);
			Saturation[i][j]=(Saturation[i][j]/64);
			Value[i][j]=(Value[i][j]/8);
			Color[i][j]=(Value[i][j]*8 + Saturation[i][j]*2 + Hue[i][j]);
        }
    }

	IplImage *imageDst;
	imageDst=cvCreateImage(cvSize(img->width,img->height),IPL_DEPTH_32F,1);
	CvScalar s;

	for(int i=0;i<imageDst->height;i++){
		for(int j=0;j<imageDst->width;j++){
			s.val[0]=Color[i][j];			
			cvSet2D(imageDst,i,j,s);
        }
    }

	img = imageDst;
	imageDst=NULL;

    SetSize(img->height,img->width);
    for(int i=0,ih=img->height,iw=img->width;i<ih;i++)
    {	
		T* pdata = p[i];
		int* pimg = reinterpret_cast<int*>(img->imageData+img->widthStep*i);
		for(int j=0;j<iw;j++){
			pdata[j] = pimg[j];
		}
    }
   
	for(int i = 0 ; i< img->height ; i++){
		delete [] Saturation[i];
		delete [] Hue[i];
		delete [] Value[i];
		delete [] Color[i];
	}
	delete [] Saturation;
	delete [] Hue;
	delete [] Value;
	delete [] Color;

	cvReleaseImage(&img);
    return true;
}

template<class T>
IplImage* IntImage<T>::convertImageRGBtoHSV(const IplImage *imageRGB){
	float fR, fG, fB;
	float fH, fS, fV;
	const float FLOAT_TO_BYTE = 255.0f;
	const float BYTE_TO_FLOAT = 1.0f / FLOAT_TO_BYTE;

	// Create a blank HSV image
	IplImage *imageHSV = cvCreateImage(cvGetSize(imageRGB), 8, 3);
	if (!imageHSV || imageRGB->depth != 8 || imageRGB->nChannels != 3) {
		printf("ERROR in convertImageRGBtoHSV()! Bad input image.\n");
		exit(1);
	}

	int h = imageRGB->height;		// Pixel height.
	int w = imageRGB->width;		// Pixel width.
	int rowSizeRGB = imageRGB->widthStep;	// Size of row in bytes, including extra padding.
	char *imRGB = imageRGB->imageData;	// Pointer to the start of the image pixels.
	int rowSizeHSV = imageHSV->widthStep;	// Size of row in bytes, including extra padding.
	char *imHSV = imageHSV->imageData;	// Pointer to the start of the image pixels.
	for (int y=0; y<h; y++) {
		for (int x=0; x<w; x++) {
			// Get the RGB pixel components. NOTE that OpenCV stores RGB pixels in B,G,R order.
			uchar *pRGB = (uchar*)(imRGB + y*rowSizeRGB + x*3);
			int bB = *(uchar*)(pRGB+0);	// Blue component
			int bG = *(uchar*)(pRGB+1);	// Green component
			int bR = *(uchar*)(pRGB+2);	// Red component

			// Convert from 8-bit integers to floats.
			fR = bR * BYTE_TO_FLOAT;
			fG = bG * BYTE_TO_FLOAT;
			fB = bB * BYTE_TO_FLOAT;

			// Convert from RGB to HSV, using float ranges 0.0 to 1.0.
			float fDelta;
			float fMin, fMax;
			int iMax;
			// Get the min and max, but use integer comparisons for slight speedup.
			if (bB < bG) {
				if (bB < bR) {
					fMin = fB;
					if (bR > bG) {
						iMax = bR;
						fMax = fR;
					}
					else {
						iMax = bG;
						fMax = fG;
					}
				}
				else {
					fMin = fR;
					fMax = fG;
					iMax = bG;
				}
			}
			else {
				if (bG < bR) {
					fMin = fG;
					if (bB > bR) {
						fMax = fB;
						iMax = bB;
					}
					else {
						fMax = fR;
						iMax = bR;
					}
				}
				else {
					fMin = fR;
					fMax = fB;
					iMax = bB;
				}
			}
			fDelta = fMax - fMin;
			fV = fMax;				// Value (Brightness).
			if (iMax != 0) {			// Make sure its not pure black.
				fS = fDelta / fMax;		// Saturation.
				float ANGLE_TO_UNIT = 1.0f / (6.0f * fDelta);	// Make the Hues between 0.0 to 1.0 instead of 6.0
				if (iMax == bR) {		// between yellow and magenta.
					fH = (fG - fB) * ANGLE_TO_UNIT;
				}
				else if (iMax == bG) {		// between cyan and yellow.
					fH = (2.0f/6.0f) + ( fB - fR ) * ANGLE_TO_UNIT;
				}
				else {				// between magenta and cyan.
					fH = (4.0f/6.0f) + ( fR - fG ) * ANGLE_TO_UNIT;
				}
				// Wrap outlier Hues around the circle.
				if (fH < 0.0f)
					fH += 1.0f;
				if (fH >= 1.0f)
					fH -= 1.0f;
			}
			else {
				// color is pure Black.
				fS = 0;
				fH = 0;	// undefined hue
			}

			// Convert from floats to 8-bit integers.
			int bH = (int)(0.5f + fH * 255.0f);
			int bS = (int)(0.5f + fS * 255.0f);
			int bV = (int)(0.5f + fV * 255.0f);

			// Clip the values to make sure it fits within the 8bits.
			if (bH > 255)
				bH = 255;
			if (bH < 0)
				bH = 0;
			if (bS > 255)
				bS = 255;
			if (bS < 0)
				bS = 0;
			if (bV > 255)
				bV = 255;
			if (bV < 0)
				bV = 0;

			// Set the HSV pixel components.
			uchar *pHSV = (uchar*)(imHSV + y*rowSizeHSV + x*3);
			*(pHSV+0) = bH;		// H component
			*(pHSV+1) = bS;		// S component
			*(pHSV+2) = bV;		// V component
		}
	}
	return imageHSV;
}

template<class T>
void IntImage<T>::Save(const std::string& filename) const
{
    IplImage* img;

    img = cvCreateImage(cvSize(ncol,nrow),IPL_DEPTH_8U,1);
    for(int i=0,ih=img->height,iw=img->width;i<ih;i++)
    {
        T* pdata = p[i];
        unsigned char* pimg = reinterpret_cast<unsigned char*>(img->imageData+img->widthStep*i);
        for(int j=0;j<iw;j++) pimg[j] = (unsigned char)pdata[j];
    }
    cvSaveImage(filename.c_str(),img);
    cvReleaseImage(&img);
}

template<class T>
void IntImage<T>::SetSize(const int h,const int w)
// 'size' is the new size of the image, if necessary, memory is reallocated
// size.cx is the new height and size.cy is the new width
{
    if((h == nrow) && (w == ncol)) return;

    Clear();
    Array2dC<T>::Create(h,w);
}

template<class T>
IntImage<T>& IntImage<T>::operator=(const IntImage<T>& source)
{
        if(&source==this) return *this;
    SetSize(source.nrow,source.ncol);
    memcpy(buf,source.buf,sizeof(T)*nrow*ncol);
    label = source.label;
    variance = source.variance;

    return *this;
}

template<class T>
void IntImage<T>::Print(std::ostream& of) const
{
    Array2dC<T>::Print();
    of<<"\tvariance="<<variance<<",    label="<<label<<std::endl;
}

template<class T>
void IntImage<T>::Resize(IntImage<T> &result,const REAL ratio) const
{
    Resize(result,int(nrow*ratio),int(ncol*ratio));
}

template<class T>
void IntImage<T>::Resize(IntImage<T>& result,const int height,const int width) const
{
    assert(height>0 && width>0);
    result.SetSize(height,width);
    REAL ixratio = nrow*1.0/height, iyratio = ncol*1.0/width;

    REAL* p_y = new REAL[result.ncol]; assert(p_y!=NULL);
    int* p_y0 = new int[result.ncol]; assert(p_y0!=NULL);
    for(int i=0;i<width;i++)
    {
        p_y[i] = i*iyratio;
        p_y0[i] = (int)p_y[i];
        if(p_y0[i]==ncol-1) p_y0[i]--;
        p_y[i] -= p_y0[i];
    }

    for(int i=0;i<height;i++)
    {
        int x0; REAL x;
        x = i*ixratio;
        x0 = (int)x;
        if(x0==nrow-1) x0--;
        x -= x0;
        for(int j=0;j<width;j++)
        {
            int y0=p_y0[j];
            REAL y=p_y[j],fx0,fx1;

            fx0 = REAL(p[x0][y0] + y*(p[x0][y0+1]-p[x0][y0]));
            fx1 = REAL(p[x0+1][y0] + y*(p[x0+1][y0+1]-p[x0+1][y0]));

            result.p[i][j] = T(fx0 + x*(fx1-fx0));
        }
    }

    delete[] p_y; p_y=NULL;
    delete[] p_y0; p_y0=NULL;
}

template<class T>
void IntImage<T>::Swap(IntImage<T>& image2)
{
    Array2dC<T>::Swap(image2);
    std::swap(variance,image2.variance);
    std::swap(label,image2.label);
}

template<class T>
void IntImage<T>::operator-=(const IntImage<T>& img2)
{
    assert(nrow==img2.nrow && ncol==img2.ncol);
    for(int i=0,size=nrow*ncol;i<size;i++) p[i] -= img2.p[i];
}

template<class T>
void IntImage<T>::AbsoluteValue(void)
{
    for(int i=0,size=nrow*ncol;i<size;i++)  if(buf[i]<0) buf[i] = -buf[i];
}

template<class T>
void IntImage<T>::Thresh(const REAL thresh)
{
    for(int i=0,size=nrow*ncol;i<size;i++) buf[i]=(buf[i]>thresh)?REAL(1.0):0;
}

#endif //__INTIMAGE_H__


